from enum import Enum
from typing import AsyncGenerator

from langgraph.checkpoint.postgres.aio import AsyncPostgresSaver
from langgraph.prebuilt import create_react_agent
from langchain_core.messages import HumanMessage, BaseMessage, \
    SystemMessage, ToolMessage, AIMessage, AIMessageChunk

from langchain_neo4j import GraphCypherQAChain, Neo4jGraph
from langchain.chains import FalkorDBQAChain
from langchain_community.graphs import FalkorDBGraph

from ..databases.postgres import Database
from ..models import ChatModel
from ..utils.logger import Logger

import os

from langchain_mcp_adapters.client import MultiServerMCPClient
from datetime import datetime
from langchain.tools import Tool

from langfuse.callback import CallbackHandler


PROMPT_MESSAGE = """You are an Airflow chatbot designed to assist Data Engineers, Data Analysts, and other team members. By leveraging Airflow MCP and a variety of other tools, you help teams navigate data more effectively and streamline their workflows."""


class LLMEventType(Enum):
    """Event types for the LLM agent."""

    STORED_MESSAGE = 'stored_message'
    RETRIEVER_START = 'on_retriever_start'
    RETRIEVER_END = 'on_retriever_end'
    CHAT_CHUNK = 'on_chat_model_stream'
    DONE = 'done'


class ChatMessage:
    class Sender(Enum):
        """The sender of the message."""
        SYSTEM = 'system'
        AI = 'ai'
        HUMAN = 'human'
        TOOL = 'tool'

    def __init__(self, type: LLMEventType, sender: Sender,
                 content: str, payload: dict = None):
        self.type = type
        self.sender: str = sender.value
        self.content = content
        self.payload = payload or {}

    @classmethod
    def from_base_message(cls, message: BaseMessage) -> 'ChatMessage':
        message_type_lookup = {
            HumanMessage: cls.Sender.HUMAN,
            SystemMessage: cls.Sender.SYSTEM,
            ToolMessage: cls.Sender.TOOL,
            AIMessage: cls.Sender.AI,
            AIMessageChunk: cls.Sender.AI,
        }

        # Different message types have different structures.
        try:
            content = message.content[0]['text']
        except (KeyError, TypeError, IndexError):
            content = message.content

        return ChatMessage(
            LLMEventType.STORED_MESSAGE,
            sender=message_type_lookup[type(message)],
            content=content,
        )

    @classmethod
    def from_event(cls, event: dict) -> 'ChatMessage':
        """Convert an event from the LLM agent to a `ChatMessage` object."""
        if event['event'] in ('on_tool_start', 'on_tool_end'):
            print(event)
            print('--------------------')
        match event['event']:
            case 'on_chat_model_stream':
                if event['data']['chunk'].content:
                    return cls._handle_on_chat_model_stream(event)
            case 'on_tool_start':
                return ChatMessage(LLMEventType.CHAT_CHUNK, cls.Sender.AI,
                                   f'''\n\nStart Running Tool:
```
Data: {event['data']['input']}
Function: {event['name']}
```
''')
            case 'on_tool_end':
                return ChatMessage(LLMEventType.CHAT_CHUNK, cls.Sender.AI,
                                   f'''\n\nTool Output: \n```\n
{event['data']['output'].content}
\n```\n''')
            # The conversation is done.
            case 'done':
                return ChatMessage(
                    LLMEventType.DONE,
                    cls.Sender.SYSTEM,
                    'Done',
                )
            # Known events that we ignore.
            case 'on_chat_model_start' | 'on_chain_start' | 'on_chain_end' \
                | 'on_chat_model_stream' | 'on_chat_model_end' | \
                    'on_chain_stream' | 'on_tool_start' | 'on_tool_end':
                Logger().get_logger().debug('Ignoring message', event['event'])
                return ''
            # Unknown events.
            case _:
                raise ValueError('Unknown event', event)

    @classmethod
    def _handle_on_chat_model_stream(cls, event: dict) -> 'ChatMessage':
        content = event['data']['chunk'].content
        content_type = ''
        if not isinstance(content, str):
            content_type = content[0]['type']
            content = content[0].get('text')

        # If the message is a tool call, just print a debug message.
        if content_type in ('tool_use', 'tool_call'):
            Logger().get_logger().debug('Stream.tool_calls:',
                                        event['data']['chunk'].tool_calls,
                                        flush=True)
            return ''
        else:
            return ChatMessage(LLMEventType.CHAT_CHUNK, cls.Sender.AI, content
                               if content is not None else '')

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the message."""
        return {
            'sender': self.sender,
            'content': self.content,
            'payload': self.payload,
        }


class LLMAgent:

    def __init__(self, tools, md_uri=None):
        self._agent = None
        self._llm = None
        self.retriever_tool_name = 'Internal_Company_Info_Retriever'
        self._checkpointer_ctx = None
        self.tools = tools
        self.md_uri = md_uri

    async def __aenter__(self) -> 'LLMAgent':

        tools = self.tools
        self._llm = ChatModel()

        # Checkpointer for the agent.
        self._checkpointer_ctx = AsyncPostgresSaver\
            .from_conn_string(
                Database(self.md_uri).get_connection_string())
        checkpointer = await self._checkpointer_ctx.__aenter__()

        # Create the agent itself.
        self._agent = create_react_agent(
            self._llm,
            tools,
            checkpointer=checkpointer,
            # state_modifier=SystemMessage(PROMPT_MESSAGE),
            prompt=PROMPT_MESSAGE
        )

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Close the agent and the checkpointer."""
        await self._checkpointer_ctx.__aexit__(exc_type, exc_val, exc_tb)
        self._llm = None
        self._agent = None
        self._checkpointer_ctx = None

    async def astream_events(self, message: str,
                             chat_session: dict) -> AsyncGenerator[ChatMessage,
                                                                   None]:
        async for event in self._agent.astream_events(
            {"messages": [HumanMessage(content=message)]},
            config=chat_session,
            version='v2',
        ):
            message = ChatMessage.from_event(event)
            if message:
                yield message

        # Let the client know that the conversation is done.
        # yield ChatMessage.from_event({'event': 'done'})


def get_user_chat_config(session_id: str, username: str = None) -> dict:
    chat_config = {'configurable': {'thread_id': session_id},
                   "recursion_limit": 100}
    if os.environ.get('LANGFUSE_HOST'):
        langfuse_handler = CallbackHandler(
                user_id=username if username else session_id,
                session_id=f"{session_id}",
                public_key=os.environ.get('LANGFUSE_PUBLIC_KEY'),
                secret_key=os.environ.get('LANGFUSE_SECRET_KEY'),
                host=os.environ.get('LANGFUSE_HOST')
            )
        chat_config['callbacks'] = [langfuse_handler]
    return chat_config


async def get_stream_agent_responce(session_id, message,
                                    md_uri: str = None,
                                    username: str = None):
    user_config = get_user_chat_config(session_id, username)
    mcp_host = os.environ.get('mcp_host', 'mcp_sse_server:8000')
    TRANSPORT_TYPE = os.environ.get('TRANSPORT_TYPE', 'stdio')
    if TRANSPORT_TYPE == 'stdio':
        mcps = {
                "AirflowMCP":
                {
                    'command': "python",
                    'args': ["-m", "airflow_mcp_hipposys.mcp_airflow"],
                    "transport": "stdio",
                    'env': {k: v for k, v in {
                        'AIRFLOW_ASSISTENT_AI_CONN': os.getenv(
                            'AIRFLOW_ASSISTENT_AI_CONN'),
                        'airflow_api_url': os.getenv('airflow_api_url'),
                        'airflow_username': os.getenv('airflow_username'),
                        'airflow_password': os.getenv('airflow_password'),
                        'AIRFLOW_INSIGHTS_MODE':
                            os.getenv('AIRFLOW_INSIGHTS_MODE'),
                        'POST_MODE': os.getenv('POST_MODE'),
                        'TRANSPORT_TYPE': 'stdio',
                        '_AIRFLOW_WWW_USER_USERNAME':
                            os.getenv('_AIRFLOW_WWW_USER_USERNAME'),
                        '_AIRFLOW_WWW_USER_PASSWORD':
                            os.getenv('_AIRFLOW_WWW_USER_PASSWORD')
                    }.items() if v is not None}
                }
            }
    elif TRANSPORT_TYPE == 'sse':
        mcps = {
                    "AirflowMCP": {
                        "url": f"http://{mcp_host}/sse",
                        "transport": "sse",
                        "headers": {"Authorization": f"""Bearer {
                            os.environ.get('MCP_TOKEN')}"""}
                    }
                }
    datetime_tool = Tool(
        name="Datetime",
        func=lambda x: datetime.now().isoformat(),
        description="Returns the current datetime",
    )

    client = MultiServerMCPClient(mcps)
    tools = await client.get_tools() + [datetime_tool]
    dbt_prompt = '''Don't use this tool to receieve charts/dashboards/Superset/datasets/datasource metadata. This graph doesn't know anything about them. Find real database name from Superset MCP.
    The database schema includes:
- Node Types (labels): "Macro", "Model" (not dataset!!!!, dataset must be discovered in Superset API first of all and search by object name), "Operation", "Seed", "Snapshot", "Source" (source's name is "schema"."name" unlike all the other nodes where it's just "name"), "Test"
- Model attributes:
    access:	e.g. protected
    alias:	The actual object name in the database, which typically matches the model name but may differ in some cases.
    checksum	
    database
    description	
    enabled: true/false
    language: sql (can be python for some adapters)
    materialized: table, view, etc.
    meta
    name:	model name (not always dataset/datasource name from Superset!!! So always retrieve Dataset/datasource metadata before it from SupersetMCP)
    original_file_path	models/{{path}}/{{name}}.sql
    owner
    package_name
    path	{{path}}/{{name}}.sql
    relation_name	"{{database}}"."{{schema}}"."{{alias}}"
    resource_type	model
    schema	
    table_type	
    tags	[]
    unique_id	model.{{package_name}}.{{name}}
- Relationships:
    DEPENDS_ON: This relationship connects nodes with the "Model", "Snapshot" label to nodes with the "Model"/"Snapshot"/"Seed"/"Source" label
    REFERENCES: This relationship connects nodes with the "Model", "Snapshot" label to nodes with the "Model"/"Snapshot"/"Seed" label (not "Source")
    TESTS: This relationship connects nodes with the "Test" label to nodes with the "Model"/"Snapshot"/"Source"/"Seed" label
    USES_MACRO: This relationship connects nodes with the "Model"/"Snapshot" label to nodes with the "Macro" label
Use this retriever to answer questions about model lineage, dependencies, testing coverage, and data flow in our dbt project.
    '''
    if os.environ.get('GRAPH_DB') == 'falkordb':
        graph = FalkorDBGraph(host=os.environ.get('GRAPH_HOST', 'falkordb'),
                              port=6379, database="dbt_graph",
                              username=os.environ.get('GRAPH_USER'),
                              password=os.environ.get('GRAPH_PASSWORD'
                                                      ),)
        chain = FalkorDBQAChain.from_llm(ChatModel(), graph=graph, verbose=True, allow_dangerous_requests=True)
        retriever_tool = chain.as_tool(name="Falkor_Knowledge_Graph_Retriever_DBT",
                                            description=f"Query and retrieve dbt data from your Falkor graph database using Cypher syntax\n{dbt_prompt}")
        tools.append(retriever_tool)
    elif os.environ.get('GRAPH_DB') == 'neo4j':
        graph = Neo4jGraph(url=f"bolt://{os.environ.get('GRAPH_HOST', 'neo4j')}:7687",
                           username=os.environ.get('GRAPH_USER'),
                           password=os.environ.get('GRAPH_PASSWORD'),
                           enhanced_schema=True)
        chain = GraphCypherQAChain.from_llm(
            ChatModel(),
            graph=graph,
            verbose=True,
            allow_dangerous_requests=True,
        )
        retriever_tool = chain.as_tool(name="Neo4J_Knowledge_Graph_Retriever_DBT",
                                       description=f"Query your Neo4j graph database using Cypher to retrieve nodes, relationships, and insights.\n{dbt_prompt}")
        tools.append(retriever_tool)

    async def stream_agent_response():
        async with LLMAgent(tools=tools, md_uri=md_uri) as llm_agent:
            async for chat_msg in llm_agent.astream_events(
                 message, user_config):
                yield chat_msg.content
    return stream_agent_response



import os

from ..utils.singleton import Singleton


class Config(metaclass=Singleton):
    """Holds "global" configuration for the application."""
    @staticmethod
    def get_deploy_env() -> str:
        """Get the current deployment environment."""
        return os.environ.get('DEPLOY_ENV', 'prod').upper()



import os

from langgraph.checkpoint.postgres.aio import AsyncPostgresSaver

from ..utils.singleton import Singleton


class Database(metaclass=Singleton):
    """Represents the main database.
    Currently, provides only the connection string to the database.
    """

    def __init__(self, md_uri: str = None):
        """Initialize the database connection."""

        super().__init__()

        self.user = os.environ.get('POSTGRES_USER', 'postgres')
        self.password = os.environ.get('POSTGRES_PASSWORD')
        self.host = os.environ.get('POSTGRES_HOSTNAME', 'postgres')
        self.port = os.environ.get('POSTGRES_PORT', 5432)
        self.database = os.environ.get('POSTGRES_DB', 'chat_db')
        auth = f'{self.user}:{self.password}'
        self.uri = \
            f'postgres://{auth}@{self.host}:{self.port}/{self.database}'
        if md_uri:
            self.uri = md_uri

    def get_connection_string(self) -> str:
        """Get a URI representation of the database connection params."""
        return self.uri

    @staticmethod
    async def setup(md_uri: str = None):
        """Setup the database."""
        async with AsyncPostgresSaver.from_conn_string(
           Database(md_uri).get_connection_string()) as saver:
            await saver.setup()



from datetime import datetime, timedelta


class Singleton(type):
    """A metaclass that implements the singleton pattern.

    This metaclass ensures that only one(*) instance of a class is created and that the instance is reused for all subsequent calls.
    
    The instance is refreshed every `MAX_INSTANCE_TTL` seconds.
    
    Usage:
    ```python
    >>> class MyClass(metaclass=Singleton):
    ...     pass
        
    >>> a = MyClass()
    >>> b = MyClass()
    >>> a is b
        True
    >>> sleep(MyClass.MAX_INSTANCE_TTL + 1)
    >>> c = MyClass()
    >>> a is c
        False
    ```
    """

    _instances = {}
    _creation_time = {}
    
    MAX_INSTANCE_TTL = timedelta(seconds=5 * 60)  # 5 minutes

    def __create_instance(cls, *args, **kwargs):
        """Create a new instance of this class."""

        cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        cls._creation_time[cls] = datetime.now()

    def __call__(cls, *args, force_recreate=False, **kwargs):
        """Create a new instance of this class if it does not exist or if it is no longer valid.
        
        If the `force_recreate` parameter is set to `True`, a new instance will be created regardless of the validity of the existing instance.
        """

        # Create a new instance if it does not exist or if `force_recreate` is set.
        if force_recreate or (cls not in cls._instances):
            cls.__create_instance(*args, **kwargs)
        # Create a new instance if the existing instance is no longer valid.
        elif (cls not in cls._creation_time) or (cls._creation_time[cls] + cls.MAX_INSTANCE_TTL < datetime.now()):
            cls.__create_instance(*args, **kwargs)

        # Return the existing instance.
        return cls._instances[cls]



import uuid

from fastapi import APIRouter, Request
from fastapi.responses import StreamingResponse
from pydantic import BaseModel
from .llm import get_stream_agent_responce


chat_router = APIRouter()


class ChatRequest(BaseModel):
    message: str


@chat_router.post("/new")
async def new_chat(request: Request):
    """Create a new chat session."""
    request.session['chat_session_id'] = f'user_{uuid.uuid4()}'
    return {'results': 'ok'}


@chat_router.post("/ask")
async def chat(
    request: Request,
    chat_request: ChatRequest,
):
    if 'chat_session_id' not in request.session:
        await new_chat(request)
    # Get the user chat configuration and the LLM agent.
    stream_agent_response = await get_stream_agent_responce(request.session[
        'chat_session_id'],
                                                      chat_request.message)

    # Return the agent's response as a stream of JSON objects.
    return StreamingResponse(stream_agent_response(),
                             media_type='application/json')



import os

from langchain_aws import ChatBedrock as BaseChatBedrock


class ChatBedrock(BaseChatBedrock):
    """A wrapper for the `langchain_aws.ChatBedrock`."""

    def __init__(self, **kwargs):
        """Initialize the `ChatBedrock` with specific configuration."""
        model_type, model_id = os.environ['LLM_MODEL_ID'].split(':', 1)
        default_kwargs = {
            'model_id': model_id,
            'region_name': os.environ.get('AWS_DEFAULT_REGION', 'us-east-1'),
            'model_kwargs': dict(temperature=0),
        }

        super().__init__(**(default_kwargs | kwargs))



import sys
import logging

from ..utils.singleton import Singleton


class Logger(metaclass=Singleton):
    default_config = {
        'name': 'RAG-App',
        'level': 'INFO',
        'format': '[%(asctime)s|%(name)s|%(levelname)s|%(processName)s:%(threadName)s|%(filename)s, '
                  'line %(lineno)s in %(funcName)s] %(message)s',
    }

    def __init__(self,
                 config_to_use: dict = None):

        config = {**Logger.default_config, **(config_to_use or {})}
        self.logger = logging.getLogger(config['name'])

        # Set the root logging level.
        logging_lvl = getattr(logging, config['level'])
        logging.root.setLevel(logging_lvl)

        # Create an handler and configure it.
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging_lvl)
        handler.setFormatter(logging.Formatter(config['format']))

        # Remove all other handlers in the process.
        self.logger.addHandler(handler)

    def get_logger(self) -> logging.Logger:
        return self.logger



from airflow import settings
from airflow.plugins_manager import AirflowPlugin
from flask import Blueprint, request, jsonify, Response, stream_template, \
    flash, url_for, redirect, current_app
from flask_login import current_user
from flask_appbuilder import expose, BaseView as AppBuilderBaseView
from functools import wraps
import json
import time
import requests
import os
from datetime import datetime, timedelta
import uuid
from sqlalchemy import Column, String, DateTime, Text, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import SQLAlchemyError
import asyncio

# Blueprint for the chat plugin
bp = Blueprint(
    "airflow_chat",
    __name__,
    template_folder="templates",
    static_folder="static",
    static_url_path="/static/airflow_chat_plugin",
)

# Add this function right after the Blueprint definition
@bp.before_app_first_request
def configure_upload_limits():
    """Configure Flask app for large file uploads"""
    current_app.config['MAX_CONTENT_LENGTH'] = 100 * 1024 * 1024  # 100MB

# Database model for storing chat sessions
Base = declarative_base()


class ChatSession(Base):
    __tablename__ = 'airflow_chat_sessions'
    
    conversation_id = Column(String(255), primary_key=True)
    cookies_data = Column(Text)  # JSON string of cookies
    created_at = Column(DateTime, default=datetime.utcnow)
    last_accessed = Column(DateTime, default=datetime.utcnow)
    
    def to_dict(self):
        return {
            'conversation_id': self.conversation_id,
            'cookies_data': self.cookies_data,
            'created_at': self.created_at.isoformat() if self.created_at else None,
            'last_accessed': self.last_accessed.isoformat() if self.last_accessed else None
        }


class ChatSessionManager:
    def __init__(self):
        # Use Airflow's metadata database
        self.engine = settings.engine
        self.Session = sessionmaker(bind=self.engine)
        
        # Create the table if it doesn't exist
        try:
            Base.metadata.create_all(self.engine)
        except Exception as e:
            print(f"Error creating chat sessions table: {e}")
    
    def cleanup_old_sessions(self):
        """Remove sessions older than 24 hours"""
        try:
            session = self.Session()
            cutoff_time = datetime.utcnow() - timedelta(minutes=20)
            
            deleted_count = session.query(ChatSession).filter(
                ChatSession.created_at < cutoff_time
            ).delete()
            
            session.commit()
            session.close()
            
            if deleted_count > 0:
                print(f"Cleaned up {deleted_count} old chat sessions")
                
        except SQLAlchemyError as e:
            print(f"Error cleaning up old sessions: {e}")
            if session:
                session.rollback()
                session.close()
    
    def store_session(self, conversation_id, cookies_dict):
        """Store or update a chat session"""
        try:
            session = self.Session()
            
            # Convert cookies to JSON string
            cookies_json = json.dumps(dict(cookies_dict)) if cookies_dict else "{}"
            
            # Check if session exists
            existing_session = session.query(ChatSession).filter(
                ChatSession.conversation_id == conversation_id
            ).first()
            
            if existing_session:
                # Update existing session
                existing_session.cookies_data = cookies_json
                existing_session.last_accessed = datetime.utcnow()
            else:
                # Create new session
                new_session = ChatSession(
                    conversation_id=conversation_id,
                    cookies_data=cookies_json,
                    created_at=datetime.utcnow(),
                    last_accessed=datetime.utcnow()
                )
                session.add(new_session)
            
            session.commit()
            session.close()
            return True
            
        except SQLAlchemyError as e:
            print(f"Error storing session: {e}")
            if session:
                session.rollback()
                session.close()
            return False
    
    def get_session(self, conversation_id):
        """Retrieve a chat session"""
        try:
            session = self.Session()
            
            chat_session = session.query(ChatSession).filter(
                ChatSession.conversation_id == conversation_id
            ).first()
            
            if chat_session:
                # Update last accessed time
                chat_session.last_accessed = datetime.utcnow()
                session.commit()
                
                # Parse cookies from JSON
                cookies_dict = json.loads(chat_session.cookies_data) if chat_session.cookies_data else {}
                
                session.close()
                return cookies_dict
            else:
                session.close()
                return None
                
        except SQLAlchemyError as e:
            print(f"Error retrieving session: {e}")
            if session:
                session.close()
            return None
    
    def delete_session(self, conversation_id):
        """Delete a specific chat session"""
        try:
            session = self.Session()
            
            deleted_count = session.query(ChatSession).filter(
                ChatSession.conversation_id == conversation_id
            ).delete()
            
            session.commit()
            session.close()
            
            return deleted_count > 0
            
        except SQLAlchemyError as e:
            print(f"Error deleting session: {e}")
            if session:
                session.rollback()
                session.close()
            return False
    
    def get_active_sessions_count(self):
        """Get count of active sessions (within 24 hours)"""
        try:
            session = self.Session()
            cutoff_time = datetime.utcnow() - timedelta(hours=24)
            
            count = session.query(ChatSession).filter(
                ChatSession.created_at >= cutoff_time
            ).count()
            
            session.close()
            return count
            
        except SQLAlchemyError as e:
            print(f"Error getting session count: {e}")
            if session:
                session.close()
            return 0

def admin_only(f):
    """Decorator to restrict access to Admin role."""
    @wraps(f)
    def decorated_function(*args, **kwargs):
        # Check if user is authenticated
        if not current_user or not hasattr(current_user, 'roles') or not current_user.roles:
            return jsonify({"error": "Authentication required"}), 401
        
        users_roles = [role.name for role in current_user.roles]
        approved_roles = ["Admin", "AIRFLOW_AI"]  # Adjust roles as needed
        if not any(role in users_roles for role in approved_roles):
            flash("You do not have permission to access this page.", "danger")
            return redirect(url_for("Airflow.index"))
        return f(*args, **kwargs, username=current_user.username)
    return decorated_function

def failure_tolerant(f):
    """Decorator for error handling."""
    @wraps(f)
    def decorated_function(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            print(f"CHAT PLUGIN FAILURE: {e}")
            return jsonify({"error": "Something went wrong"}), 500
    return decorated_function

class LLMChatAgent:
    def __init__(self):
        self.backend_url = os.environ.get('BACKEND_URL', "http://fastapi:8080")
        self.access_token = os.environ.get('FAST_API_ACCESS_SECRET_TOKEN', 'ThisIsATempAccessTokenForLocalEnvs.ReplaceInProd')
        self.session_manager = ChatSessionManager()
        
        # Perform cleanup on initialization
        self.session_manager.cleanup_old_sessions()
    
    def get_headers(self):
        return {
            "x-access-token": self.access_token
        }
    
    def initialize_chat_session(self, conversation_id):
        """Initialize a new chat session"""
        try:
            response = requests.post(
                f"{self.backend_url}/chat/new",
                headers=self.get_headers(),
                timeout=10
            )
            response.raise_for_status()
            
            # Store session cookies in database
            success = self.session_manager.store_session(conversation_id, response.cookies)
            
            if success:
                print(f"Initialized and stored chat session: {conversation_id}")
                return True, response.cookies
            else:
                print(f"Failed to store session in database: {conversation_id}")
                return False, None
            
        except requests.RequestException as e:
            print(f"Failed to initialize chat session: {e}")
            return False, e
        except Exception as e:
            print(f"Unexpected error initializing session: {e}")
            return False, e
    
    def stream_chat_response(self, message, conversation_id=None):
        """Stream response from FastAPI backend"""
        try:
            # Clean up old sessions periodically (every request)
            self.session_manager.cleanup_old_sessions()
            
            # Get session cookies from database
            cookies_dict = self.session_manager.get_session(conversation_id)
            
            if cookies_dict is None:
                # Initialize session if not found
                success, cookies = self.initialize_chat_session(conversation_id)
                if not success:
                    yield {
                        "content": "Failed to initialize chat session. Please try again.",
                        "conversation_id": conversation_id,
                        "timestamp": datetime.now().isoformat(),
                        "error": True
                    }
                    return
                cookies_dict = dict(cookies)
            
            # Convert dict back to requests.cookies.RequestsCookieJar if needed
            # For requests library, we can pass cookies as dict
            
            # Stream chat response
            with requests.post(
                f"{self.backend_url}/chat/ask",
                stream=True,
                headers=self.get_headers(),
                json={'message': message},
                cookies=cookies_dict,
                timeout=30
            ) as response:
                
                response.raise_for_status()
                
                # Stream response chunks
                for chunk in response.iter_content(chunk_size=1024):
                    if chunk:
                        try:
                            # Decode the chunk
                            content = str(chunk, encoding="utf-8")
                            
                            # Yield the content as structured response
                            yield {
                                "content": content,
                                "conversation_id": conversation_id,
                                "timestamp": datetime.now().isoformat()
                            }
                            
                        except UnicodeDecodeError:
                            # Skip chunks that can't be decoded
                            continue
                        except Exception as e:
                            print(f"Error processing chunk: {e}")
                            continue
                
        except requests.RequestException as e:
            # Fallback for connection errors
            yield {
                "content": f"Connection error: {str(e)}. Please check if the FastAPI service is running.",
                "conversation_id": conversation_id,
                "timestamp": datetime.now().isoformat(),
                "error": True
            }
        except Exception as e:
            yield {
                "content": f"An unexpected error occurred: {str(e)}",
                "conversation_id": conversation_id,
                "timestamp": datetime.now().isoformat(),
                "error": True
            }
    
    def clear_session(self, conversation_id):
        """Clear session data for a conversation"""
        print(f'Clearing session: {conversation_id}')
        return self.session_manager.delete_session(conversation_id)
    
    def get_session_stats(self):
        """Get statistics about active sessions"""
        return {
            'active_sessions': self.session_manager.get_active_sessions_count()
        }

class DBTMetadataLoader:
    """Handler for DBT metadata loading functionality"""
    
    @staticmethod
    def is_graph_db_enabled():
        """Check if GRAPH_DB environment variable is set"""
        return bool(os.environ.get('GRAPH_DB'))
    
    @staticmethod
    def load_dbt_metadata(manifest_content, catalog_content):
        """Load DBT metadata to the configured graph database"""
        graph_db = os.environ.get('GRAPH_DB')
        
        if graph_db == 'falkordb':
            from dbt_graph_loader.loaders.falkordb_loader import DBTFalkorDBLoader
            loader = DBTFalkorDBLoader(
                host=os.environ.get("GRAPH_HOST", "falkordb"),
                username=os.environ.get('GRAPH_USER'),
                password=os.environ.get('GRAPH_PASSWORD')
            )
            loader.load_dbt_to_falkordb_from_strings(manifest_content, catalog_content)
            
        elif graph_db == 'neo4j':
            from dbt_graph_loader.loaders.neo4j_loader import DBTNeo4jLoader
            loader = DBTNeo4jLoader(
                f'neo4j://{os.environ.get("GRAPH_HOST", "neo4j")}:7687',
                os.environ.get('GRAPH_USER'),
                os.environ.get('GRAPH_PASSWORD')
            )
            loader.load_dbt_to_neo4j_from_strings(manifest_content, catalog_content)
            
        else:
            raise ValueError(f'Unsupported GRAPH_DB value: {graph_db}. Supported values are: falkordb, neo4j')

class AirflowChatView(AppBuilderBaseView):
    default_view = "chat_interface"
    route_base = "/airflow_chat"
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.llm_agent = LLMChatAgent()
        self.dbt_loader = DBTMetadataLoader()
    
    @expose("/")
    @admin_only
    @failure_tolerant
    def chat_interface(self, **kwargs):
        """Main chat interface"""
        return self.render_template("chat_interface.html", 
                                   graph_db_enabled=self.dbt_loader.is_graph_db_enabled())
    
    @expose("/dbt_upload")
    @admin_only
    @failure_tolerant
    def dbt_upload_interface(self, **kwargs):
        """DBT metadata upload interface (only shown if GRAPH_DB is set)"""
        if not self.dbt_loader.is_graph_db_enabled():
            flash("DBT metadata upload is not available. GRAPH_DB environment variable is not set.", "warning")
            return redirect(url_for("AirflowChatView.chat_interface"))
        
        return self.render_template("dbt_upload.html", 
                                   graph_db=os.environ.get('GRAPH_DB'))
    
    @expose("/api/chat", methods=["POST"])
    @admin_only
    @failure_tolerant
    def chat_api(self, **kwargs):
        """API endpoint for chat messages"""
        username = kwargs.get('username')
        data = request.get_json()
        message = data.get("message", "").strip()
        conversation_id = data.get("conversation_id")
        print(f'conversation_id: {conversation_id}')
        
        if not message:
            return jsonify({"error": "Message is required"}), 400
        
        # Return streaming response
        def generate(username=None):
            try:
                if str(os.environ.get('INTERNAL_AI_ASSISTANT_SERVER', True))\
                   .lower() == 'true':
                    from .app.server.llm import get_stream_agent_responce
                    md_uri = str(settings.Session().bind.url).replace('postgresql+psycopg2', 'postgres')
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
                    try:
                        # Get the async generator function
                        stream_agent_response = loop.run_until_complete(
                            get_stream_agent_responce(conversation_id, message, md_uri, username)
                        )
                        
                        # Stream each chunk as it comes
                        async_gen = stream_agent_response()
                        while True:
                            try:
                                chunk = loop.run_until_complete(async_gen.__anext__())
                                # Ensure chunk has the same structure as the else branch
                                if isinstance(chunk, dict):
                                    # If chunk is already a dict, use it as is
                                    formatted_chunk = chunk
                                else:
                                    # If chunk is raw content, wrap it in the expected format
                                    formatted_chunk = {
                                        "content": str(chunk),
                                        "conversation_id": conversation_id or str(uuid.uuid4()),
                                        "timestamp": datetime.now().isoformat(),
                                        "error": False
                                    }
                                yield f"data: {json.dumps(formatted_chunk)}\n\n"
                            except StopAsyncIteration:
                                break
                        
                        yield "data: [DONE]\n\n"
                        
                    finally:
                        loop.close()
                else:
                    for chunk in self.llm_agent.stream_chat_response(
                        message,
                        conversation_id
                    ):
                        yield f"data: {json.dumps(chunk)}\n\n"
                    yield "data: [DONE]\n\n"
            except Exception as e:
                error_chunk = {
                    "content": f"Error: {str(e)}",
                    "conversation_id": conversation_id or str(uuid.uuid4()),
                    "timestamp": datetime.now().isoformat(),
                    "error": True
                }
                yield f"data: {json.dumps(error_chunk)}\n\n"
                yield "data: [DONE]\n\n"
        
        return Response(
            generate(username=username),
            mimetype="text/plain",
            headers={
                "Cache-Control": "no-cache",
                "Connection": "keep-alive",
                "Access-Control-Allow-Origin": "*"
            }
        )

    @expose("/api/new_chat", methods=["POST"])
    @admin_only
    @failure_tolerant
    def new_chat(self, **kwargs):
        """Initialize a new chat session"""
        conversation_id = str(uuid.uuid4())
        
        try:
            if not str(os.environ.get('INTERNAL_AI_ASSISTANT_SERVER', True))\
               .lower() == 'true':
                success, cookies = self.llm_agent.initialize_chat_session(conversation_id)
            else:
                from .app.databases.postgres import Database
                md_uri = str(settings.Session().bind.url).replace('postgresql+psycopg2', 'postgres')
                asyncio.run(Database.setup(md_uri))
                print('Database setup complete')
                success = True
            
            if success:
                return jsonify({
                    "conversation_id": conversation_id,
                    "status": "initialized",
                    "message": "New chat session created"
                })
            else:
                return jsonify({
                    "error": f"Failed to initialize chat session: {cookies}"
                }), 500
                
        except Exception as e:
            return jsonify({
                "error": f"Error creating new chat: {str(e)}"
            }), 500
    
    @expose("/api/clear_session", methods=["POST"])
    @admin_only
    @failure_tolerant
    def clear_session(self, **kwargs):
        """Clear a specific chat session"""
        data = request.get_json()
        conversation_id = data.get("conversation_id")
        
        if not conversation_id:
            return jsonify({"error": "Conversation ID is required"}), 400
        
        try:
            success = self.llm_agent.clear_session(conversation_id)
            
            if success:
                return jsonify({
                    "message": "Session cleared successfully",
                    "conversation_id": conversation_id
                })
            else:
                return jsonify({
                    "error": "Session not found or could not be cleared"
                }), 404
                
        except Exception as e:
            return jsonify({
                "error": f"Error clearing session: {str(e)}"
            }), 500
    
    @expose("/api/session_stats", methods=["GET"])
    @admin_only
    @failure_tolerant
    def session_stats(self, **kwargs):
        """Get session statistics"""
        try:
            stats = self.llm_agent.get_session_stats()
            return jsonify(stats)
        except Exception as e:
            return jsonify({
                "error": f"Error getting session stats: {str(e)}"
            }), 500
    
    @expose("/api/dbt_upload", methods=["POST"])
    @admin_only
    @failure_tolerant
    def upload_dbt_metadata(self, **kwargs):
        """Upload DBT metadata files (manifest.json and catalog.json)"""
        if not self.dbt_loader.is_graph_db_enabled():
            return jsonify({
                "error": "DBT metadata upload is not available. GRAPH_DB environment variable is not set."
            }), 400
        
        try:
            # Check request size before processing
            content_length = request.content_length
            max_size = 100 * 1024 * 1024  # 100MB
            
            if content_length and content_length > max_size:
                return jsonify({
                    "error": f"Request too large. Maximum size is {max_size / 1024 / 1024}MB"
                }), 413
            
            # Check if files are present in the request
            if 'manifest_file' not in request.files or 'catalog_file' not in request.files:
                return jsonify({
                    "error": "Both manifest_file and catalog_file are required"
                }), 400
            
            manifest_file = request.files['manifest_file']
            catalog_file = request.files['catalog_file']
            
            # Validate file names
            if manifest_file.filename == '' or catalog_file.filename == '':
                return jsonify({
                    "error": "Both files must have valid filenames"
                }), 400
            
            # Validate file extensions
            if not manifest_file.filename.endswith('.json') or not catalog_file.filename.endswith('.json'):
                return jsonify({
                    "error": "Both files must be JSON files"
                }), 400
            
            # Read file contents
            manifest_content = manifest_file.read()
            catalog_content = catalog_file.read()
            
            # Validate JSON format
            try:
                json.loads(manifest_content)
                json.loads(catalog_content)
            except json.JSONDecodeError as e:
                return jsonify({
                    "error": f"Invalid JSON format: {str(e)}"
                }), 400
            
            # Load metadata to graph database
            self.dbt_loader.load_dbt_metadata(manifest_content, catalog_content)
            
            return jsonify({
                "message": "DBT metadata uploaded successfully",
                "graph_db": os.environ.get('GRAPH_DB'),
                "manifest_file": manifest_file.filename,
                "catalog_file": catalog_file.filename
            })
            
        except ImportError as e:
            return jsonify({
                "error": f"Required DBT graph loader library not found: {str(e)}"
            }), 500
        except ValueError as e:
            return jsonify({
                "error": str(e)
            }), 400
        except Exception as e:
            return jsonify({
                "error": f"Error uploading DBT metadata: {str(e)}"
            }), 500
    
    @expose("/api/graph_db_status", methods=["GET"])
    @admin_only
    @failure_tolerant
    def graph_db_status(self, **kwargs):
        """Get the current graph database configuration status"""
        return jsonify({
            "graph_db_enabled": self.dbt_loader.is_graph_db_enabled(),
            "graph_db": os.environ.get('GRAPH_DB'),
            "graph_user": bool(os.environ.get('GRAPH_USER')),
            "graph_password": bool(os.environ.get('GRAPH_PASSWORD'))
        })

# Create the view instance
chat_view = AirflowChatView()

# Package for Airflow
chat_package = {
    "name": "AI Chat Assistant",
    "category": "Tools",
    "view": chat_view,
}

# Conditionally add DBT Upload menu item only if GRAPH_DB is set
menu_items = []


class AirflowChatPlugin(AirflowPlugin):
    name = "airflow_chat_plugin"
    hooks = []
    macros = []
    flask_blueprints = [bp]
    appbuilder_views = [chat_package]
    appbuilder_menu_items = []


import os
model_type, model_id = os.environ.get('LLM_MODEL_ID', 'none:none').split(':', 1)

if model_type == 'bedrock':
    from .inference.bedrock_model import ChatBedrock
    ChatModel = ChatBedrock
elif model_type == 'antropic':
    from .inference.antropic_model import ChatAnthropic
    ChatModel = ChatAnthropic
elif model_type == 'openai':
    from .inference.openai_model import ChatOpenAI
    ChatModel = ChatOpenAI
else:
    ChatModel = None



import os

from contextlib import asynccontextmanager
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse
from starlette.middleware.sessions import SessionMiddleware

from .chat import chat_router
from ..databases.postgres import Database
from ..utils.config import Config
from ..utils.logger import Logger


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Run the database setup and teardown."""
    await Database.setup()
    Logger().get_logger().info('Database setup complete')

    yield
    # Optionally add teardown code here


app = FastAPI(lifespan=lifespan)

app.add_middleware(
    SessionMiddleware,
    secret_key=os.environ.get('SECRET_KEY'),
    https_only=Config.get_deploy_env() == 'PROD',
)


@app.middleware("http")
async def check_token_middleware(request: Request, call_next):
    """Allow only requests with the correct token."""
    token = request.headers.get("x-access-token")
    if (Config.get_deploy_env() != 'LOCAL') \
       and (token != os.environ['FAST_API_ACCESS_SECRET_TOKEN']):
        return JSONResponse(
            status_code=403, content={'reason': 'Invalid or missing token'})
    response = await call_next(request)
    return response


app.include_router(chat_router, prefix='/chat')



import os

from langchain_anthropic import ChatAnthropic as BaseChatAnthropic


class ChatAnthropic(BaseChatAnthropic):
    """A wrapper for the `langchain_aws.ChatBedrock`."""

    def __init__(self, **kwargs):
        """Initialize the `ChatBedrock` with specific configuration."""
        model_type, model_id = os.environ['LLM_MODEL_ID'].split(':', 1)
        default_kwargs = {
            'model': model_id,
            'temperature': 0,
        }

        super().__init__(**(default_kwargs | kwargs))



import os

from langchain_openai import ChatOpenAI as BaseChatOpenAI


class ChatOpenAI(BaseChatOpenAI):
    """A wrapper for the `langchain_aws.ChatOpenAI`."""

    def __init__(self, model_id: str = None, **kwargs):
        """Initialize the `ChatOpenAI` with specific configuration."""
        model_type, model_id_env = os.environ['LLM_MODEL_ID'].split(':', 1)
        default_kwargs = {
            'model': model_id or model_id_env,
            'temperature': 0,
        }

        super().__init__(**(default_kwargs | kwargs))
